import config
import subprocess
import os
import serial
import threading
import json
from bottle import route, run
import time
import sys
import signal
import logging

logging.basicConfig(filename='pixelserver.log',level=config.LogLevel)

running = True

if config.UseGui:
    import pygame

    class Gui(threading.Thread):
        def __init__(self):
            super().__init__()
            self.data = b"\x00"*config.ScreenX*config.ScreenY*3
            self.lock = threading.Lock()

        def setData(self, data):
            with self.lock:
                self.data = data

        def run(self):
            global running
            logging.info("Starting GUI")
            sf = config.GuiScaleFactor
            screen = pygame.display.set_mode((sf*config.ScreenX, sf*config.ScreenY))
            pygame.display.set_caption("Pixelserver - GUI Vis")
            while running:
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        logging.debug("GUI quits app")
                        running = False
                screen.fill((0, 255, 0))
                with self.lock:
                    try:
                        for x in range(config.ScreenX):
                            for y in range(config.ScreenY):
                                i = x+y*config.ScreenX
                                r = (self.data[i*3+0])
                                g = (self.data[i*3+1])
                                b = (self.data[i*3+2])

                                pygame.draw.rect(screen, (r, g, b), pygame.Rect(sf*x, sf*y, sf, sf))
                    except:
                        pass
                time.sleep(1/30)
                pygame.display.flip()
            logging.info("Closing GUI")

class SerialWriter(threading.Thread):
    def __init__(self):
        super().__init__()
        self.lock = threading.Lock()
        self.data = b"\x00"*config.ScreenX*config.ScreenY*3

    def setData(self, data):
        with self.lock:
            self.data = data
    
    def run(self):
        should_connect = True
        ser = None
        logging.info("Starting SerialWriter")
        while running:
            try:
                if should_connect:
                    ser = serial.Serial(config.Serial)
                    should_connect = False
                    logging.info("Serial Opened")
                with self.lock:
                    ser.write(b"\01")
                    ser.write(self.data)
                time.sleep(1/120)
            except Exception as e:
                if ser != None:
                    ser.close()
                    ser = None
                logging.warning("Serial was close because: "+str(e))
                should_connect = True
                time.sleep(0.1)
        logging.info("Closing SerialWriter")

class WatchDog(threading.Thread):
    def __init__(self, check, action):
        super().__init__()
        self.check = check
        self.action = action
    def run(self):
        while running:
            if self.check():
                logging.error("Watchdog: Executed")
                self.action()
            time.sleep(1)


class LogReader(threading.Thread):
    def __init__(self, runner):
        super().__init__()
        self.runner = runner
        self.log = ""
    def clear(self):
        self.log = ""

    def run(self):
        logging.info("LogReader started")
        while running:
            try:
                self.log += runner.app.stderr.read(1).decode("utf-8")
            except Exception as e:
                time.sleep(0.1) 
        logging.info("LogReader closed")


class AppRunner(threading.Thread):
    def __init__(self):
        super().__init__()
        self.currentApp = -1
        self.requestedApp = 0
        self.app = None
        self.lock = threading.Lock()
        self.param = ""
        self.serial = SerialWriter()
        self.serial.start()
        self.last_update = time.time()
        self.watchdog = WatchDog(lambda: self.appTimedOut(), lambda: self.terminateApp())
        self.watchdog.start()
        self.logreader = LogReader(self)
        self.logreader.start()
        if config.UseGui:
            self.gui = Gui()
            self.gui.start()

    def requestApp(self, app, param=""):
        with self.lock:
            self.requestedApp = app
            self.param = param

    def terminateApp(self):
        if self.app != None:
            logging.error("Terminate app "+config.Apps[self.currentApp]["name"])
            # don't ask
            self.app.terminate()
        if self.app != None:
            self.app.kill()
            self.app.stdout.close()
            self.app.stderr.close()
            self.last_update = time.time()
        self.requestApp = 0

    def appTimedOut(self):
        return time.time()-self.last_update > config.NoDataTimeout

    def updateApp(self):
        if self.app != None:
            self.app.terminate()
            self.app.stderr.close()
        cmd = config.Apps[self.requestedApp]["cmd"]
        logging.debug(str(cmd))
        if type(cmd) != list:
            cmd = [cmd,]
        args = cmd+[str(config.ScreenX), str(config.ScreenY), self.param]
        self.app = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        self.logreader.clear()

    def run(self):
        logging.info("Starting Apprunner")
        while running:
            with self.lock:
                if self.app == None or self.app.poll() != None:
                    self.requestedApp = 0   
                if self.requestedApp != None:
                    self.currentApp = self.requestedApp
                    self.updateApp()
                    self.requestedApp = None
            d = self.app.stdout
            oshandle = d.fileno()
            try:
                data = os.read(oshandle, config.ScreenX*config.ScreenY*3)
                self.last_update = time.time()
                self.serial.setData(data)
                if config.UseGui:
                    self.gui.setData(data)
            except:
                logging.debug("Exception in Apprunner.run")
        self.watchdog.join()
        self.serial.join()
        logreader.join()
        if config.UseGui:
            self.gui.join()
        logging.info("Close Apprunner")

    def getLog(self):
        return self.logreader.log

runner = AppRunner()
runner.start()

@route("/apps/list")
def apps_list():
    s = []
    for app in config.Apps:
        s.append(app["name"])
    return json.dumps(s)
    
@route("/apps/start/<name>")
def apps_start(name):
    for i in range(len(config.Apps)):
        if config.Apps[i]["name"] == name:
            runner.requestApp(i)
            return "ok"
    return "not_found"

@route("/apps/start/<name>/<param>")
def apps_start(name, param):
    for i in range(len(config.Apps)):

        if config.Apps[i]["name"] == name:
            runner.requestApp(i, param)
            return "ok"
    return "not_found"

@route("/apps/log")
def apps_log():
    return runner.getLog()

@route("/apps/running") 
def  apps_running():
    i = runner.currentApp
    return config.Apps[i]["name"]

@route("/")
def index():
    return open("html/index.html").read()

run(host="localhost", port=8000)
running = False
runner.join()
